
/**
 *
 * \file     kmath_random.cpp
 * \brief    generate random number
 * \authoers Toshiyuki Imamura (TI)
 * \date     2013/12/10 (NT)
 * 
 * (c) Copyright 2013 RIKEN. All rights reserved.
 */

#include "kmath_random.h"
#include "../dsfmt/km_dsfmt_jump.h"

#include <stdlib.h>


KMATH_Random::KMATH_Random()
{
  _isInit    = false;
  _comm      = 0;
  _comm_rank = 0;
  _comm_size = 0;
  _dsfmt     = new dsfmt_t;
}

KMATH_Random::~KMATH_Random()
{
  finalize();

  delete (dsfmt_t*)_dsfmt;
}

bool
KMATH_Random::init(MPI_Comm comm)
{
  if (_isInit)
    return false;

  if (MPI_Comm_dup(comm, &_comm) != 0)
    return false;

  if (MPI_Comm_rank(_comm, &_comm_rank) != 0 ||
      MPI_Comm_size(_comm, &_comm_size) != 0)
    {
      MPI_Comm_free(&_comm);
      return false;
    }

  int seed = 1;
  int ret;
  km_dsfmt_jump_do_((dsfmt_t*)_dsfmt, &seed, &_comm_rank, &_comm_size, &ret);
  if (ret != 0)
    {
      MPI_Comm_free(&_comm);
      return false;
    }

  _isInit = true;
  return true;
}

bool
KMATH_Random::finalize()
{
  if (!_isInit)
    return false;

  MPI_Comm_free(&_comm);

  _isInit = false;
  return true;
}

bool
KMATH_Random::seed(int seed)
{
  if (!_isInit)
    return false;

  int ret;
  km_dsfmt_jump_do_((dsfmt_t*)_dsfmt, &seed, &_comm_rank, &_comm_size, &ret);

  return ret == 0;
}

bool
KMATH_Random::get(double& value) const
{
  if (!_isInit)
    return false;

  value = dsfmt_genrand_close1_open2((dsfmt_t*)_dsfmt);
  return true;
}

bool
KMATH_Random::get(double* value, int size) const
{
  if (!_isInit)
    return false;

  if (size < 384 || size % 2 != 0)
    return false;

  dsfmt_fill_array_close1_open2((dsfmt_t*)_dsfmt, value, size);
  return true;
}

int
KMATH_Random::serialize(const char* filename)
{
  if (!_isInit || filename == NULL)
    return -1;

  int size = sizeof(dsfmt_t) * _comm_size;
  unsigned char* buf = new unsigned char[size];

  int ret = 0;

  if (MPI_Gather(_dsfmt, sizeof(dsfmt_t), MPI_BYTE,
		 buf,    sizeof(dsfmt_t), MPI_BYTE,
		 0, _comm) != MPI_SUCCESS)
    ret = -2;

  if (_comm_rank == 0 && ret == 0)
    {
      FILE* fp = fopen(filename, "wb");
      if (fp == NULL)
	ret = -3;

      if (ret == 0 && fwrite(&_comm_size, sizeof(int), 1, fp) != 1)
	ret = -3;

      if (ret == 0 && fwrite(buf, size, 1, fp) != 1)
	ret = -3;

      if (fp != NULL)
	fclose(fp);
    }

  if (MPI_Bcast(&ret, 1, MPI_INTEGER, 0, _comm) != MPI_SUCCESS)
    ret = -2;

  delete buf;
  return ret;
}

int
KMATH_Random::deserialize(const char* filename)
{
  if (!_isInit || filename == NULL)
    return -1;

  int size = sizeof(dsfmt_t) * _comm_size;
  unsigned char* buf = new unsigned char[size];

  int ret = 0;

  if (_comm_rank == 0)
    {
      FILE* fp = fopen(filename, "rb");
      if (fp == NULL)
	ret = -3;

      int nrank;
      if (ret == 0 && fread(&nrank, sizeof(int), 1, fp) != 1)
	ret = -3;

      if (ret == 0 && _comm_size != nrank)
	ret = -4;

      if (ret == 0 && fread(buf, size, 1, fp) != 1)
	ret = -3;

      if (fp != NULL)
	fclose(fp);
    }

  if (MPI_Bcast  (&ret, 1, MPI_INTEGER, 0, _comm) != MPI_SUCCESS)
    ret = -2;

  if (MPI_Scatter(buf,    sizeof(dsfmt_t), MPI_BYTE,
		  _dsfmt, sizeof(dsfmt_t), MPI_BYTE,
		  0, _comm) != MPI_SUCCESS)
    ret = -2;

  delete buf;
  return ret;
}
