
/**
 *
 * \file     kmath_random.c
 * \brief    generate random number
 * \authoers Toshiyuki Imamura (TI)
 * \date     2013/02/04 (NT)
 * \date     2013/12/10 (NT)
 * 
 * (c) Copyright 2013 RIKEN. All rights reserved.
 */

#include "kmath_random.h"
#include "../dsfmt/km_dsfmt_jump.h"

#include <stdlib.h>


typedef struct _Context
{
  MPI_Comm comm;
  int      comm_rank;
  int      comm_size;
  dsfmt_t  dsfmt;

} Context;


inline void*
KMATH_Random_init(MPI_Comm comm)
{
  Context* c = (Context*)malloc(sizeof(Context));

  if (MPI_Comm_dup(comm, &c->comm) != 0)
    {
      free(c);
      return NULL;
    }

  if (MPI_Comm_rank(c->comm, &c->comm_rank) != 0 ||
      MPI_Comm_size(c->comm, &c->comm_size) != 0)
    {
      MPI_Comm_free(&c->comm);
      free(c);
      return NULL;
    }

  int seed = 1;
  int ret;
  km_dsfmt_jump_do_(&c->dsfmt, &seed, &c->comm_rank, &c->comm_size, &ret);
  if (ret != 0)
    {
      MPI_Comm_free(&c->comm);
      free(c);
      return NULL;
    }

  return (void*)c;
}


inline int
KMATH_Random_finalize(void* handle)
{
  if (handle == NULL)
    return -1;

  Context* c = (Context *)handle;

  MPI_Comm_free(&c->comm);

  free(c);

  return 0;
}


inline int
KMATH_Random_seed(void* handle, int seed)
{
  if (handle == NULL)
    return -1;

  Context* c = (Context *)handle;

  int ret;
  km_dsfmt_jump_do_(&c->dsfmt, &seed, &c->comm_rank, &c->comm_size, &ret);

  return ret;
}


inline int
KMATH_Random_get(void* handle, double* value)
{
  if (handle == NULL)
    return -1;

  Context* c = (Context *)handle;

  *value = dsfmt_genrand_close1_open2(&c->dsfmt);
  return 0;
}


inline int
KMATH_Random_vector(void* handle, double* value, int size)
{
  if (handle == NULL)
    return -1;

  if (size < 384 || size % 2 != 0)
    return -2;

  Context* c = (Context *)handle;

  dsfmt_fill_array_close1_open2(&c->dsfmt, value, size);
  return 0;
}

int
KMATH_Random_serialize(void* handle, const char* filename)
{
  if (handle == NULL || filename == NULL)
    return -1;

  Context* c = (Context *)handle;

  int size = sizeof(dsfmt_t) * c->comm_size;
  void* buf = malloc(size);

  int ret = 0;

  if (MPI_Gather(&c->dsfmt, sizeof(dsfmt_t), MPI_BYTE,
		 buf,       sizeof(dsfmt_t), MPI_BYTE, 
		 0, c->comm) != MPI_SUCCESS)
    ret = -2;

  if (c->comm_rank == 0 && ret == 0)
    {
      FILE* fp = fopen(filename, "wb");
      if (fp == NULL)
	ret = -3;

      if (ret == 0 && fwrite(&c->comm_size, sizeof(int), 1, fp) != 1)
	ret = -3;

      if (ret == 0 && fwrite(buf, size, 1, fp) != 1)
	ret = -3;

      if (fp != NULL)
	fclose(fp);
    }

  if (MPI_Bcast(&ret, 1, MPI_INTEGER, 0, c->comm) != MPI_SUCCESS)
    ret = -2;

  free(buf);
  return ret;
}

int
KMATH_Random_deserialize(void* handle, const char* filename)
{
  if (handle == NULL || filename == NULL)
    return -1;

  Context* c = (Context *)handle;

  int size = sizeof(dsfmt_t) * c->comm_size;
  void* buf = malloc(size);

  int ret = 0;

  if (c->comm_rank == 0)
    {
      FILE* fp = fopen(filename, "rb");
      if (fp == NULL)
	ret = -3;

      int nrank;
      if (ret == 0 && fread(&nrank, sizeof(int), 1, fp) != 1)
	ret = -3;

      if (ret == 0 && c->comm_size != nrank)
	ret = -4;

      if (ret == 0 && fread(buf, size, 1, fp) != 1)
	ret = -3;

      if (fp != NULL)
        fclose(fp);
    }

  if (MPI_Bcast  (&ret, 1, MPI_INTEGER, 0, c->comm) != MPI_SUCCESS)
    ret = -2;

  if (MPI_Scatter(buf,       sizeof(dsfmt_t), MPI_BYTE,
		  &c->dsfmt, sizeof(dsfmt_t), MPI_BYTE,
		  0, c->comm) != MPI_SUCCESS)
    ret = -2;

  free(buf);
  return ret;
}
