
!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_random_mod
!> @brief   random number module
!! @authors Toshiyuki Imamura (TI)
!! @date    2013/02/06 (NT)
!! @date    2013/12/10 (NT)
!
!  (c) Copyright 2013 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_random_mod

  use mpi

  implicit none
  private


  ! public procedures
  !

  public :: KMATH_Random_Init
  public :: KMATH_Random_Finalize
  public :: KMATH_Random_Seed
  public :: KMATH_Random_Get
  public :: KMATH_Random_Vector
  public :: KMATH_Random_Serialize
  public :: KMATH_Random_Deserialize


  ! global variables
  !

  type, public:: s_km_rand

    logical            :: init = .false.
    integer            :: comm
    integer            :: comm_rank
    integer            :: comm_size
    character, pointer :: dsfmt(:)

  end type s_km_rand


  ! private procedures
  !
  private :: get_free_unit

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Init
  !> @brief      Initialize random number system
  !! @authors    NT
  !! @param[out] handle    handle for this random number
  !! @param[in]  comm      communicator
  !! @param[out] ierr      error information (0: no error)
  !! @date       2013/12/10 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Init(handle, comm, ierr)

    ! formal arguments
    type(s_km_rand),  target, intent(inout) :: handle
    integer,                  intent(in)    :: comm
    integer,                  intent(out)   :: ierr

    ! local variables
    type(s_km_rand),  pointer :: h
    integer                   :: ierr2, isize, seed


    h => handle

    ! check handle initialization
    !
    if (h%init) then
      ierr = -1
      return 
    end if


    ! duplicate main communicator
    !
    call MPI_Comm_dup(comm, h%comm, ierr)
    if (ierr /= 0) &
      return 


    ! check the nranks
    !
    call MPI_Comm_size(h%comm, h%comm_size, ierr)
    if (ierr /= 0) then
      call MPI_Comm_free(h%comm, ierr2)
      return 
    end if

    call MPI_Comm_rank(h%comm, h%comm_rank, ierr)
    if (ierr /= 0) then
      call MPI_Comm_free(h%comm, ierr2)
      return
    end if


    ! allocate dSFMT working memory
    !
    call km_size_of_dsfmt_t(isize)

    allocate(h%dsfmt(isize))


    ! initialize random seed
    !
    seed = 1

    call km_dsfmt_jump_do(h%dsfmt, seed, h%comm_rank, h%comm_size, ierr)
    if (ierr /= 0) then
      call MPI_Comm_free(h%comm, ierr2)
      return
    end if


    h%init = .true.
    return

  end subroutine KMATH_Random_Init


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Finalize
  !> @brief      Finalize random number system
  !! @authors    NT
  !! @param[in]  handle   handle for this random number
  !! @param[out] ierr     error information (0: no error)
  !! @date       2013/02/06 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Finalize(handle, ierr)

    ! formal arguments
    type(s_km_rand),  target, intent(inout) :: handle
    integer,                  intent(out)   :: ierr

    ! local variables
    type(s_km_rand),  pointer :: h


    h => handle

    if (.not. h%init) then
      ierr = -1
      return 
    end if


    ! deallocate dSFMT working memory
    !
    deallocate(h%dsfmt)


    ! free MPI communicator
    !
    call MPI_Comm_free(h%comm, ierr)


    h%init = .false.

    return

  end subroutine KMATH_Random_Finalize


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Seed
  !> @brief      Setup random seed
  !! @authors    NT
  !! @param[in]  handle   handle for this random number
  !! @param[in]  seed     seed value
  !! @param[out] ierr     error information (0: no error)
  !! @date       2013/12/10 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Seed(handle, seed, ierr)

    ! formal arguments
    type(s_km_rand),  target, intent(in)  :: handle
    integer,                  intent(in)  :: seed
    integer,                  intent(out) :: ierr

    ! local variables
    type(s_km_rand),  pointer :: h


    h => handle

    if (.not. h%init) then
      ierr = -1
      return
    end if

    call km_dsfmt_jump_do(h%dsfmt, seed, h%comm_rank, h%comm_size, ierr)

    return

  end subroutine KMATH_Random_Seed


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Get
  !> @brief      Get a random value
  !! @authors    NT
  !! @param[in]  handle   handle for this random number
  !! @param[out] value    random value
  !! @param[out] ierr     error information (0: no error)
  !! @date       2013/02/06 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Get(handle, value, ierr)

    ! formal arguments
    type(s_km_rand),  intent(in)  :: handle
    double precision, intent(out) :: value
    integer,          intent(out) :: ierr


    if (.not. handle%init) then
      ierr = -1
      return 
    end if

    call km_dsfmt_genrand_close1_open2(handle%dsfmt, value)

    ierr = 0
    return

  end subroutine KMATH_Random_Get


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Vector
  !> @brief      Get random values
  !! @authors    NT
  !! @param[in]  handle   handle for this random number
  !! @param[out] values   random values
  !! @param[in]  nvalue   number of values
  !! @param[out] ierr     error information (0: no error)
  !! @date       2013/02/06 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Vector(handle, value, nvalue, ierr)

    ! formal arguments
    type(s_km_rand),  intent(in)    :: handle
    double precision, intent(inout) :: value(:)
    integer,          intent(in)    :: nvalue
    integer,          intent(out)   :: ierr


    if (.not. handle%init) then
      ierr = -1
      return 
    end if

    call km_dsfmt_fill_array_close1_open2(handle%dsfmt, value, nvalue)

    ierr = 0
    return

  end subroutine KMATH_Random_Vector


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Serialize
  !> @brief      Serialize random number system
  !! @authors    NT
  !! @param[in]  handle    handle for this random number
  !! @param[in]  filename  output file name
  !! @param[out] ierr      error information (0: no error)
  !! @date       2013/12/10 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Serialize(handle, filename, ierr)

    ! formal arguments
    type(s_km_rand),  intent(in)    :: handle
    character(*),     intent(in)    :: filename
    integer,          intent(out)   :: ierr

    ! local variables
    integer           :: dsize, asize, unit, open_stat, stat, ie

    character, allocatable :: buf(:)


    if (.not. handle%init) then
      ierr = -1
      return 
    end if

    call km_size_of_dsfmt_t(dsize)
    asize = dsize * handle%comm_size + 4

    allocate(buf(asize))

    ierr = 0

    call MPI_Gather(handle%dsfmt, dsize, MPI_BYTE, &
                    buf(5:),      dsize, MPI_BYTE, 0, handle%comm, ie)
    if (ie /= MPI_SUCCESS) &
      ierr = -2

    if (handle%comm_rank == 0 .and. ierr == 0) then

      buf(1:4) = transfer(handle%comm_size, buf(1:4))

      unit = get_free_unit()

      open(unit, file   = filename,      &
                 form   = 'unformatted', &
                 access = 'direct',      &
                 status = 'replace',     &
                 recl   = asize,         &
                 iostat = open_stat)
      if (open_stat /= 0) ierr = -3
                 
      if (ierr == 0) then
        write(unit,rec=1,iostat=stat) buf(1:asize)
        if (stat /= 0) ierr = -3
      end if

      if (open_stat == 0) &
        close(unit)

    end if

    call MPI_Bcast  (ierr, 1, MPI_INTEGER, 0, handle%comm, ie)
    if (ie /= MPI_SUCCESS) &
      ierr = -2

    deallocate(buf)

    return

  end subroutine KMATH_Random_Serialize


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Random_Deserialize
  !> @brief      Deserialize random number system
  !! @authors    NT
  !! @param[in]  handle    handle for this random number
  !! @param[in]  filename  input file name
  !! @param[out] ierr      error information (0: no error)
  !! @date       2013/12/10 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Random_Deserialize(handle, filename, ierr)

    ! formal arguments
    type(s_km_rand),  intent(inout) :: handle
    character(*),     intent(in)    :: filename
    integer,          intent(out)   :: ierr

    ! local variables
    integer           :: dsize, asize, unit, open_stat, stat, nrank, ie

    character, allocatable :: buf(:)


    if (.not. handle%init) then
      ierr = -1
      return 
    end if

    call km_size_of_dsfmt_t(dsize)
    asize = dsize * handle%comm_size + 4

    allocate(buf(asize))

    ierr = 0

    if (handle%comm_rank == 0) then

      unit = get_free_unit()

      open(unit, file   = filename,      &
                 form   = 'unformatted', &
                 access = 'direct',      &
                 status = 'old',         &
                 recl   = asize,         &
                 iostat = open_stat)
      if (open_stat /= 0) ierr = -3

      if (ierr == 0) then
        read (unit,rec=1,iostat=stat) buf(1:asize)
        if (stat /= 0) ierr = -3
      end if

      if (ierr == 0) then
        nrank = transfer(buf, nrank)
        if (nrank /= handle%comm_size) ierr = -4
      end if

      if (open_stat == 0) &
        close(unit)

    end if

    call MPI_Bcast  (ierr, 1, MPI_INTEGER, 0, handle%comm, ie)
    if (ie /= MPI_SUCCESS) &
      ierr = -2

    call MPI_Scatter(buf(5:),      dsize, MPI_BYTE, &
                     handle%dsfmt, dsize, MPI_BYTE, 0, handle%comm, ie)
    if (ie /= MPI_SUCCESS) &
      ierr = -2

    deallocate(buf)
    return

  end subroutine KMATH_Random_Deserialize


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function    get_free_unit
  !> @brief      return free unit number
  !! @authors    NT
  !! @date       2013/08/20 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  function get_free_unit()

    ! return value
    integer       :: get_free_unit

    ! local variables
    logical       :: op


    do get_free_unit = 100, 10000
      inquire(get_free_unit, opened=op)
      if (.not. op) &
        return
    end do

    get_free_unit = 0
    return

  end function get_free_unit


end module kmath_random_mod
